import re
from collections import OrderedDict
import itertools
import mistune
from mistune.renderers.markdown import MarkdownRenderer
import lxml.etree as ET


def unfold(iterable, headerfun):
	sections = []
	stack = [sections]

	for item in iterable:

		if (level := headerfun(item)) is not None:
			current_level = len(stack) - 1

			if current_level >= level:
				drop = current_level - level + 1
				stack = stack[:-drop]

			stack[-1].append([])
			stack.append(stack[-1][-1])
			for _ in range(current_level, level - 1):
				stack[-1].append([])
				stack.append(stack[-1][-1])

		stack[-1].append(item)

	return sections


class PlainRenderer(MarkdownRenderer):
	def __init__(self):
		super(PlainRenderer, self).__init__()

	def heading(self, token, state):
		return self.render_children(token, state)

	def emphasis(self, token, state):
		return self.render_children(token, state)

	def strong(self, token, state):
		return self.render_children(token, state)

	def codespan(self, token, state):
		return token['raw']

	def softbreak(self, token, state):
		return ' '


class Notes(object):
	_VERSION_RE = re.compile(r'((\d+)(\.\d+)*)$')
	_TITLE_LIST = ['Fixed Bugs and Malfunctions', 'Improvements and New Features', 'Known Bugs and Problems']

	@staticmethod
	def from_md(source):
		markdown = mistune.create_markdown(renderer=None)
		renderer = PlainRenderer()
		state = mistune.core.BlockState()

		ast = markdown(source.read().decode("utf-8"))
		ast = unfold(ast, lambda tok: tok['attrs']['level'] if tok['type'] == 'heading' else None)

		islist = lambda tok: isinstance(tok, list)
		sections = next(filter(islist, ast))

		change_list = []
		for section in filter(islist, sections):
			heading = renderer.render_children(section[0], state)
			version_match = Notes._VERSION_RE.search(heading)
			if not version_match:
				continue

			version = version_match.group(0)
			changes = []

			for subsection in filter(islist, section):
				heading = renderer.render_children(subsection[0], state)
				if not heading in Notes._TITLE_LIST:
					continue

				for tok in filter(lambda tok: tok['type'] == 'list', subsection):

					for item in tok['children']:
						if not ('children' in item and len(item['children'])):
							continue

						text = renderer([item['children'][0]], state).strip()
						changes.append(text)

			change_list.append((version, list(changes)))

		return Notes(change_list)

	@staticmethod
	def from_xml(source, parser=None):
		if parser is None:
			parser = ET.XMLParser()

		root = ET.parse(source, parser=parser)
		change_list = []

		for x in root.findall('./section[title]'):
			version_match = Notes._VERSION_RE.search(x.find("title").text)
			if not version_match:
				continue

			version = version_match.group(0)
			entries = []
			for item in itertools.chain.from_iterable([x.findall("./section[title='{}']/list/item[p]".format(title)) for title in Notes._TITLE_LIST]):
				entries.append(" ".join(itertools.chain.from_iterable(["".join(y.itertext()).split() for y in item.xpath("./p[position()<last()]")])))

			change_list.append((version, entries))

		return Notes(change_list)

	def __init__(self, changelog):
		versions, self._changes = zip(*changelog)
		self._versions = OrderedDict(((ver, idx) for idx, ver in enumerate(versions)))

	@property
	def app_versions(self):
		return self._versions.keys()

	def __getitem__(self, version):
		return self._changes[self._versions[version]]

	def slice(self, version_begin, version_end):
		idx_begin = self._versions[version_begin] if version_begin else None
		idx_end = self._versions[version_end] if version_end else None

		return itertools.islice(self._changes, idx_end, idx_begin)



import datetime
import glob
import itertools
import os
import os.path
import textwrap

DEFAULT_AUTHOR = 'opensuse-packaging@opensuse.org'

class Changes(object):
	wrapper = textwrap.TextWrapper(width=67, initial_indent='  * ', subsequent_indent='    ')

	@staticmethod
	def find_changes(path=None):
		if path is None:
			path = os.getcwd()
		return [os.path.split(x) for x in glob.glob(os.path.join(path, "*.changes"))]

	@staticmethod
	def from_otp_src(otp_src, base_otp_version, author=None):
		if author is None:
			author = DEFAULT_AUTHOR

		otp_version = otp_src.otp_version
		otp_versions_table = otp_src.otp_versions_table

		if base_otp_version == otp_version:
			return Changes(author, [])

		def proc_app(app, prev_ver, next_ver):
			if next_ver is None:
				return (app, ["Application {} deleted".format(app)])

			changes = itertools.chain.from_iterable(otp_src.app_notes_xml[app].slice(prev_ver, next_ver))
			return (app, list(changes))

		def proc_version(ver):
			changelog = [proc_app(app, *vpair) for (app, vpair) in otp_versions_table.diff(ver).items() if vpair[0] != vpair[1]]
			return (ver, changelog)

		changelog = [proc_version(ver) for ver in otp_versions_table.slice(base_otp_version, otp_version)]

		return Changes(author, changelog)

	def _generate_changelog(self, changelog, author):
		lines = [str('-' * 67), str("%s - %s" % (datetime.datetime.utcnow().strftime('%a %b %d %H:%M:%S UTC %Y'), author)), str("")]

		for version, apps in changelog:
			lines.append("- Changes for {}:".format(version))
			for (app, changes) in apps:
				for x in changes:
					lines.extend(Changes.wrapper.wrap(app + ": " + x))

		lines.extend(["",]*2)

		self._content = "\n".join(lines).encode("utf-8")

	def __init__(self, changelog, author):
		self._content = b""
		self._generate_changelog(author, changelog)

	def write(self, sink):
		if not hasattr(sink, "write"):
			sink = open(sink, "r+b")

		existing_content = sink.read()
		sink.seek(0)
		sink.write(self._content + existing_content)



import re
import requests


_re_refs_tags = re.compile(r'refs/tags/(.*)')
_re_OTP = re.compile(r'OTP-(.*)')


def get_github_tags(prefix, owner="erlang", repo="otp"):
	url = "https://api.github.com/repos/{}/{}/git/matching-refs/tags/{}".format(owner, repo, prefix)
	req = requests.get(url)
	req.raise_for_status()
	return [_re_refs_tags.match(x['ref']).groups(1)[0] for x in req.json()]

def get_github_version(major):
	prefix = "OTP-{}".format(major)
	return _re_OTP.match(get_github_tags(prefix)[-1]).groups(1)[0]



import codecs
import re
import rpm

class Spec(object):
	def __init__(self, filename):
		#
		# Workaround for https://github.com/rpm-software-management/rpm/pull/1067
		# See also http://lists.rpm.org/pipermail/rpm-list/2020-February/002012.html
		#
		rpm.reloadConfig()

		self._spec = rpm.spec(filename)
		self._filename = filename

	@property
	def version(self):
		version_header = self._spec.sourceHeader[rpm.RPMTAG_VERSION]
		if isinstance(version_header, (bytes, bytearray)):
			return version_header.decode("ascii")
		return version_header

	@property
	def filename(self):
		return self._filename

	@property
	def sources(self):
		return [source for source,position,flag in self._spec.sources if flag == 1]

	@staticmethod
	def _replace_tag(filename, tag, string):
		with codecs.open(filename, 'r+', 'utf8') as f:
			contents = f.read()
			f.seek(0)

			contents_new, subs = re.subn(r'^{tag}:([ \t\f\v]*)[^%\n\r]*'.format(tag=tag),
				r'{tag}:\g<1>{string}'.format(tag=tag, string=string),
				contents, flags=re.MULTILINE)

			if subs > 0:
				f.truncate()
				f.write(contents_new)

	@staticmethod
	def set_version(filename, version):
		Spec._replace_tag(filename, "Version", version)
